import itertools

import numpy as np

from stimupy.components.shapes import cross as cross_shape
from stimupy.components.shapes import rectangle as rectangle_shape
from stimupy.utils import make_two_sided, pad_dict_to_shape, resolution

__all__ = [
    "rectangle_generalized",
    "rectangle",
    "rectangle_two_sided",
    "cross_generalized",
    "cross",
    "cross_two_sided",
    "equal",
    "equal_two_sided",
]


def rectangle_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    target_size=None,
    target_position=None,
    covers_size=None,
    covers_x=None,
    covers_y=None,
    intensity_background=0.0,
    intensity_target=0.5,
    intensity_covers=1.0,
):
    """Rectangular target and rectangular covers added with flexible number of covers and flexible target and cover placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    target_size : float or (float, float)
        size of the target in degrees of visual angle (height, width)
    target_position : float or (float, float)
        coordinates where to place the target
    covers_size : float or (float, float)
        size of the covers in degrees of visual angle (height, width)
    covers_x : tuple of floats
        x coordinates of covers; as many covers as there are coordinates
    covers_y : tuple of floats
        y coordinates of covers; as many covers as there are coordinates
    intensity_background : float
        intensity value for background
    intensity_target : float
        intensity value for target
    intensity_covers : Sequence[Number, ...] or Number
        intensity value for covers

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    Todorovic, D. (1997).
        Lightness and junctions. Perception, 26, 379-395.
    """
    if target_size is None:
        raise ValueError(
            "rectangle_generalized() missing argument 'target_size' which is not 'None'"
        )
    if covers_size is None:
        raise ValueError(
            "rectangle_generalized() missing argument 'covers_size' which is not 'None'"
        )
    if covers_x is None:
        raise ValueError("rectangle_generalized() missing argument 'covers_x' which is not 'None'")
    if covers_y is None:
        raise ValueError("rectangle_generalized() missing argument 'covers_y' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if isinstance(covers_size, (float, int)):
        covers_size = (covers_size, covers_size)
    if len(covers_x) != len(covers_y):
        raise ValueError("Need as many x- as y-coordinates")

    # Create image with square
    stim = rectangle_shape(
        visual_size=visual_size,
        ppd=ppd,
        rectangle_size=target_size,
        rectangle_position=target_position,
        intensity_background=intensity_background,
        intensity_rectangle=intensity_target,
    )
    img = stim["img"]
    mask = stim["rectangle_mask"]

    # Add covers
    cheight, cwidth = resolution.lengths_from_visual_angles_ppd(
        covers_size, np.unique(ppd), round=False
    )
    cx = resolution.lengths_from_visual_angles_ppd(covers_x, np.unique(ppd), round=False)
    cy = resolution.lengths_from_visual_angles_ppd(covers_y, np.unique(ppd), round=False)

    cheight = int(np.round(cheight))
    cwidth = int(np.round(cwidth))
    cx = np.round(cx).astype(int)
    cy = np.round(cy).astype(int)

    if isinstance(intensity_covers, (float, int)):
        int_cov = [
            intensity_covers,
        ]
    else:
        int_cov = list(intensity_covers)
    int_cov = itertools.cycle(int_cov)

    for i in range(len(covers_x)):
        img[cy[i] : cy[i] + cheight, cx[i] : cx[i] + cwidth] = next(int_cov)
        mask[cy[i] : cy[i] + cheight, cx[i] : cx[i] + cwidth] = 0
        if cy[i] + cheight > shape[0] or cx[i] + cwidth > shape[1]:
            raise ValueError("Covers do not fully fit into stimulus")

    stim["img"] = img
    stim["target_mask"] = mask.astype(int)
    stim["visual_size"] = visual_size
    stim["ppd"] = ppd
    stim["shape"] = shape
    stim["target_size"] = stim["rectangle_size"]
    stim["target_position"] = stim["rectangle_position"]
    stim["intensity_target"] = stim["intensity_rectangle"]
    stim["covers_size"] = covers_size
    stim["covers_x"] = covers_x
    stim["covers_y"] = covers_y
    stim["intensity_covers"] = intensity_covers
    del stim["rectangle_size"]
    del stim["rectangle_position"]
    del stim["intensity_rectangle"]
    return stim


def rectangle(
    visual_size=None,
    ppd=None,
    shape=None,
    target_size=None,
    target_position=None,
    covers_size=None,
    covers_offset=None,
    intensity_background=0.0,
    intensity_target=0.5,
    intensity_covers=1.0,
):
    """Rectangular target in the center and four rectangular covers added symmetrically around target center

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    target_size : float or (float, float)
        size of the target in degrees of visual angle (height, width)
    target_position : float or (float, float)
        coordinates where to place the target
    covers_size : float or (float, float)
        size of covers in degrees of visual angle (height, width)
    covers_offset : float or (float, float)
        distance from cover center to target center in (y, x)
    intensity_background : float
        intensity value for background
    intensity_target : float
        intensity value for target
    intensity_covers : Sequence[Number, ...] or Number
        intensity value for covers

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    Todorovic, D. (1997).
        Lightness and junctions. Perception, 26, 379-395.
    """
    if target_size is None:
        raise ValueError("rectangle() missing argument 'target_size' which is not 'None'")
    if covers_size is None:
        raise ValueError("rectangle() missing argument 'covers_size' which is not 'None'")
    if covers_offset is None:
        raise ValueError("rectangle() missing argument 'covers_offset' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    if isinstance(covers_size, (float, int)):
        covers_size = (covers_size, covers_size)
    if isinstance(covers_offset, (float, int)):
        covers_offset = (covers_offset, covers_offset)

    # Calculate placement of covers for generalized function:
    cy, cx = np.array(visual_size) / 2
    y1 = cy - covers_offset[0] - covers_size[0] / 2
    x1 = cx - covers_offset[1] - covers_size[1] / 2
    y2 = cy + covers_offset[0] - covers_size[0] / 2
    x2 = cx + covers_offset[1] - covers_size[1] / 2

    stim = rectangle_generalized(
        visual_size=visual_size,
        ppd=ppd,
        target_size=target_size,
        target_position=target_position,
        covers_size=covers_size,
        covers_x=(x1, x2, x2, x1),
        covers_y=(y1, y2, y1, y2),
        intensity_background=intensity_background,
        intensity_target=intensity_target,
        intensity_covers=intensity_covers,
    )
    return stim


rectangle_two_sided = make_two_sided(
    rectangle,
    two_sided_params=(
        "target_size",
        "covers_size",
        "covers_offset",
        "intensity_background",
        "intensity_target",
        "intensity_covers",
    ),
)


def cross_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_size=None,
    cross_thickness=None,
    covers_size=None,
    covers_x=None,
    covers_y=None,
    intensity_background=0.0,
    intensity_target=0.5,
    intensity_covers=1.0,
):
    """Cross target and rectangular covers added with flexible number of covers and flexible cover placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cross_size : float or (float, float)
        size of target cross in visual angle
    cross_thickness : float
        thickness of target cross in visual angle
    covers_size : float or (float, float)
        size of covers in degrees visual angle (height, width)
    covers_x : tuple of floats
        x coordinates of covers; as many covers as there are coordinates
    covers_y : tuple of floats
        y coordinates of covers; as many covers as there are coordinates
    intensity_background : float
        intensity value for background
    intensity_target : float
        intensity value for target
    intensity_covers : Sequence[Number, ...] or Number
        intensity value for covers

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    Todorovic, D. (1997).
        Lightness and junctions. Perception, 26, 379-395.
    """
    if cross_size is None:
        raise ValueError("cross_generalized() missing argument 'cross_size' which is not 'None'")
    if cross_thickness is None:
        raise ValueError(
            "cross_generalized() missing argument 'cross_thickness' which is not 'None'"
        )
    if covers_size is None:
        raise ValueError("cross_generalized() missing argument 'covers_size' which is not 'None'")
    if covers_x is None:
        raise ValueError("cross_generalized() missing argument 'covers_x' which is not 'None'")
    if covers_y is None:
        raise ValueError("cross_generalized() missing argument 'covers_y' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    ppd = np.unique(ppd)[0]

    if not isinstance(ppd, (float, int)):
        raise ValueError("ppd should be equal in x and y direction")
    if isinstance(covers_x, (float, int)):
        covers_x = (covers_x,)
    if isinstance(covers_y, (float, int)):
        covers_y = (covers_y,)
    if len(covers_x) != len(covers_y):
        raise ValueError("Need as many x- as y-coordinates")
    if isinstance(covers_size, (float, int)):
        covers_size = (covers_size, covers_size)
    if isinstance(cross_size, (float, int)):
        cross_size = (cross_size, cross_size)

    if cross_size[0] < cross_thickness or cross_size[1] < cross_thickness:
        raise ValueError("cross_size needs to be larger than cross_thickness")

    stim = cross_shape(
        visual_size=cross_size,
        ppd=ppd,
        cross_size=cross_size,
        cross_arm_ratios=1,
        cross_thickness=cross_thickness,
        intensity_background=intensity_background,
        intensity_cross=intensity_target,
    )

    stim = pad_dict_to_shape(stim, shape=shape, pad_value=intensity_background)
    img = stim["img"]
    mask = stim["cross_mask"]

    cheight, cwidth = resolution.lengths_from_visual_angles_ppd(
        covers_size, np.unique(ppd), round=False
    )
    cx = resolution.lengths_from_visual_angles_ppd(covers_x, np.unique(ppd), round=False)
    cy = resolution.lengths_from_visual_angles_ppd(covers_y, np.unique(ppd), round=False)

    cheight = int(np.round(cheight))
    cwidth = int(np.round(cwidth))
    cx = np.round(cx).astype(int)
    cy = np.round(cy).astype(int)

    if isinstance(intensity_covers, (float, int)):
        int_cov = [
            intensity_covers,
        ]
    else:
        int_cov = list(intensity_covers)
    int_cov = itertools.cycle(int_cov)

    for i in range(len(covers_x)):
        img[cy[i] : cy[i] + cheight, cx[i] : cx[i] + cwidth] = next(int_cov)
        mask[cy[i] : cy[i] + cheight, cx[i] : cx[i] + cwidth] = 0
        if cy[i] + cheight > shape[0] or cx[i] + cwidth > shape[1]:
            raise ValueError("Covers do not fully fit into stimulus")

    stim["img"] = img
    stim["target_mask"] = mask.astype(int)
    stim["target_size"] = stim["cross_size"]
    stim["intensity_target"] = intensity_target
    stim["covers_size"] = covers_size
    stim["covers_x"] = covers_x
    stim["covers_y"] = covers_y
    stim["intensity_covers"] = intensity_covers
    stim["visual_size"] = visual_size
    stim["ppd"] = ppd
    stim["shape"] = shape
    del (stim["cross_size"], stim["intensity_cross"], stim["cross_mask"])
    return stim


def cross(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_size=None,
    cross_thickness=None,
    covers_size=None,
    intensity_background=0.0,
    intensity_target=0.5,
    intensity_covers=1.0,
):
    """Cross target and four rectangular covers added at inner cross corners

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cross_size : float or (float, float)
        size of target cross in visual angle
    cross_thickness : float
        thickness of target cross in visual angle
    covers_size : float or (float, float)
        size of covers in degrees of visual angle (height, width)
    intensity_background : float
        intensity value for background
    intensity_target : float
        intensity value for target
    intensity_covers : Sequence[Number, ...] or Number
        intensity value for covers

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    Todorovic, D. (1997).
        Lightness and junctions. Perception, 26, 379-395.
    """
    if cross_size is None:
        raise ValueError("cross() missing argument 'cross_size' which is not 'None'")
    if cross_thickness is None:
        raise ValueError("cross() missing argument 'cross_thickness' which is not 'None'")
    if covers_size is None:
        raise ValueError("cross() missing argument 'covers_size' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")
    if isinstance(covers_size, (float, int)):
        covers_size = (covers_size, covers_size)

    # Calculate placement of covers for generalized function:
    ppd = np.unique(ppd)[0]
    cy = np.floor(shape[0] / 2) / ppd
    cx = np.floor(shape[1] / 2) / ppd
    ct = np.ceil(cross_thickness * ppd)
    ct = (ct + (ct % 2)) / ppd
    ct_half = ct / 2

    y1 = cy - cross_thickness / 2 - covers_size[0]
    x1 = cx - cross_thickness / 2 - covers_size[1]
    y2 = cy + ct_half
    x2 = cx + ct_half

    stim = cross_generalized(
        visual_size=visual_size,
        ppd=ppd,
        cross_size=cross_size,
        cross_thickness=ct,
        covers_size=covers_size,
        covers_x=(x1, x2, x2, x1),
        covers_y=(y1, y2, y1, y2),
        intensity_background=intensity_background,
        intensity_target=intensity_target,
        intensity_covers=intensity_covers,
    )
    return stim


cross_two_sided = make_two_sided(
    cross,
    two_sided_params=(
        "cross_size",
        "cross_thickness",
        "covers_size",
        "intensity_background",
        "intensity_target",
        "intensity_covers",
    ),
)


def equal(
    visual_size=None,
    ppd=None,
    shape=None,
    cross_size=None,
    cross_thickness=None,
    intensity_background=0.0,
    intensity_target=0.5,
    intensity_covers=1.0,
):
    """Cross target and four rectangular covers added at inner cross corners

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cross_size : float or (float, float)
        size of target cross in visual angle
    cross_thickness : float
        thickness of target cross in visual angle
    intensity_background : float
        intensity value for background
    intensity_target : float
        intensity value for target
    intensity_covers : float
        intensity value for covers

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    Todorovic, D. (1997).
        Lightness and junctions. Perception, 26, 379-395.
    """
    if cross_size is None:
        raise ValueError("equal() missing argument 'cross_size' which is not 'None'")
    if cross_thickness is None:
        raise ValueError("equal() missing argument 'cross_thickness' which is not 'None'")

    if isinstance(cross_size, (float, int)):
        cross_size = (cross_size, cross_size)

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    # Calculate placement of covers for generalized function:
    ppd = np.unique(ppd)[0]
    c1 = (np.ceil(cross_size[0] / 2 * ppd - cross_thickness / 2 * ppd + 1)) / ppd
    c2 = (np.ceil(cross_size[1] / 2 * ppd - cross_thickness / 2 * ppd + 1)) / ppd
    covers_size = (c1, c2)

    # covers_size = ((cross_size[0] - cross_thickness) / 2, (cross_size[1] - cross_thickness) / 2)

    stim = cross(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        cross_size=cross_size,
        cross_thickness=cross_thickness,
        covers_size=covers_size,
        intensity_background=intensity_background,
        intensity_target=intensity_target,
        intensity_covers=intensity_covers,
    )

    window = rectangle_shape(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rectangle_size=cross_size,
    )

    stim["img"] = np.where(window["rectangle_mask"], stim["img"], intensity_background)
    stim["target_mask"] = np.where(window["rectangle_mask"], stim["target_mask"], 0).astype(int)
    return stim


equal_two_sided = make_two_sided(
    equal,
    two_sided_params=(
        "cross_size",
        "cross_thickness",
        "intensity_background",
        "intensity_target",
        "intensity_covers",
    ),
)


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    rectangle_params = {
        "target_size": 3,
        "covers_size": 1.5,
    }

    cross_params = {
        "cross_size": 4,
        "cross_thickness": 1.5,
    }
    # fmt: off
    stimuli = {
        "todorovic_rectangle": rectangle(**default_params, **rectangle_params, covers_offset=1.5),
        "todorovic_rectangle_general": rectangle_generalized(**default_params, **rectangle_params, target_position=3.5, covers_x=(2, 6), covers_y=(2, 6)),
        "todorovic_rectangle_2sided": rectangle_two_sided(**default_params, **rectangle_params, covers_offset=1, intensity_background=(0.0, 1.0), intensity_covers=(1.0, 0.0)),
        "todorovic_cross": cross(**default_params, **cross_params, covers_size=2),
        "todorovic_cross_general": cross_generalized(**default_params, **cross_params, covers_size=2, covers_x=(2, 6), covers_y=(2, 6)),
        "todorovic_cross_2sided": cross_two_sided(**default_params, **cross_params, covers_size=1, intensity_background=(0.0, 1.0), intensity_covers=(1.0, 0.0)),
        "todorovic_equal": equal(**default_params, **cross_params,),
        "todorovic_equal_2sided": equal_two_sided(**default_params, **cross_params, intensity_background=(0.0, 1.0), intensity_covers=(1.0, 0.0)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



from stimupy.stimuli import rings
from stimupy.utils import make_two_sided

__all__ = [
    "circular",
    "circular_generalized",
    "circular_two_sided",
    "rectangular",
    "rectangular_generalized",
    "rectangular_two_sided",
]


def circular(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_rings=None,
    ring_width=None,
    phase_shift=0,
    intensity_target=0.5,
    intensity_rings=(0.0, 1.0),
    intensity_background=0.5,
    origin="mean",
    clip=True,
):
    """Circular Bullseye stimulus

    Circular grating, where the target is the central disc.
    Alias for circular_white(target_indices=0,...)

    Specification of the number of rings, and their width can be done in two ways:
    a ring_width (in degrees) and n_rings, and/or by specifying the spatial frequency
    of a circular grating (in cycles per degree)

    The total shape (in pixels) and visual size (in degrees) has to match the
    specification of the rings and their widths.
    Thus, not all 6 parameters (visual_size, ppd, shape, frequency, ring_width, n_rings)
    have to be specified, as long as both the resolution, and the distribution of rings,
    can be resolved.

    Note: all rings in a grating have the same width -- if more control is required
    see disc_and_rings

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of circular grating, in cycles per degree
    n_rings : int, or None (default)
        number of rings
    ring_width : Number, or None (default)
        width of a single ring, in degrees
    phase_shift : float
        phase shift of grating in degrees
    intensity_target : float (optional)
        intensity value of target ring(s), by default 0.5
    intensity_rings : Sequence[Number, ...]
        intensity value for each ring, from inside to out, by default [1,0]
        If fewer intensities are passed than number of radii, cycles through intensities
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    clip : Bool
        if True, clip stimulus to image size (default: True)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Bindman, D., & Chubb, C. (2004).
        Brightness assimilation in bullseye displays.
        Vision Research, 44, 309-319.
        https://doi.org/10.1016/S0042-6989(03)00430-9
    Hong, S. W., and Shevell, S. K. (2004).
        Brightness contrast and assimilation from patterned inducing backgrounds.
        Vision Research, 44, 35-43.
        https://doi.org/10.1016/j.visres.2003.07.010
    Howe, P. D. L. (2005).
        White's effect:
        removing the junctions but preserving the strength of the illusion.
        Perception, 34, 557-564.
        https://doi.org/10.1068/p5414
    """
    stim = rings.circular(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_rings=n_rings,
        ring_width=ring_width,
        phase_shift=phase_shift,
        intensity_rings=intensity_rings,
        intensity_background=intensity_background,
        intensity_target=intensity_target,
        target_indices=1,
        origin=origin,
        clip=clip,
    )
    return stim


def circular_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    intensity_rings=(0.0, 1.0),
    intensity_background=0.5,
    intensity_target=0.5,
    origin="mean",
):
    """Draw sequential set of circular rings with specified radii, with central target

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number] or None (default)
        radii of each ring, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_rings : Sequence[float, float]
        intensities of rings, by default (1.0, 0.0)
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each frame (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    stim = rings.circular_generalized(
        radii=radii,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        intensity_rings=intensity_rings,
        intensity_background=intensity_background,
        target_indices=1,
        intensity_target=intensity_target,
        origin=origin,
    )
    return stim


circular_two_sided = make_two_sided(
    circular,
    two_sided_params=("intensity_rings", "intensity_target"),
)


def rectangular(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_frames=None,
    frame_width=None,
    rotation=0.0,
    phase_shift=0,
    intensity_frames=(0.0, 1.0),
    intensity_background=0.5,
    intensity_target=0.5,
    origin="mean",
    clip=True,
):
    """Square "bullseye", i.e., set of rings with target in center

    Essentially frames(target_indices=1)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_frames : int, or None (default)
        number of frames in the grating
    frame_width : Number, or None (default)
        width of a single frame, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_frames : Sequence[float, float]
        min and max intensity of square-wave, by default (0.0, 1.0)
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    clip : Bool
        if True, clip stimulus to image size (default: True)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Bindman, D., & Chubb, C. (2004).
        Brightness assimilation in bullseye displays.
        Vision Research, 44, 309-319.
        https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    stim = rings.rectangular(
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        frequency=frequency,
        n_frames=n_frames,
        frame_width=frame_width,
        rotation=rotation,
        phase_shift=phase_shift,
        intensity_frames=intensity_frames,
        target_indices=1,
        intensity_target=intensity_target,
        origin=origin,
        clip=clip,
        intensity_background=intensity_background,
    )
    return stim


def rectangular_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    rotation=0.0,
    intensity_frames=(0.0, 1.0),
    intensity_background=0.5,
    intensity_target=0.5,
    origin="mean",
):
    """Draw sequential set of square frames with specified radii with central target

    Parameters
    ----------
    frame_radii : Sequence[Number]
        radii of each frame, in degrees visual angle
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number] or None (default)
        radii of each frame, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_frames : Sequence[float, float]
        min and max intensity of square-wave, by default (0.0, 1.0)
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each frame (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    stim = rings.rectangular_generalized(
        radii=radii,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        intensity_frames=intensity_frames,
        intensity_background=intensity_background,
        target_indices=1,
        intensity_target=intensity_target,
        origin=origin,
    )
    return stim


rectangular_two_sided = make_two_sided(
    rectangular,
    two_sided_params=("intensity_frames", "intensity_target"),
)


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "bullseyes_circular": circular(**default_params, frequency=1.0, clip=True),
        "bullseyes_circular_2sided": circular_two_sided(**default_params, frequency=1.0, intensity_rings=((0.0, 1.0), (1.0, 0.0))),
        "bullseyes_rectangular": rectangular(**default_params, frequency=1.0, clip=True),
        "bullseyes_rectangular_2sided": rectangular_two_sided(**default_params, frequency=1.0, intensity_frames=((0.0, 1.0), (1.0, 0.0))),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import itertools
import warnings
from copy import deepcopy

import numpy as np

from stimupy.components import draw_regions, image_base
from stimupy.utils import int_factorize, resolution
from stimupy.utils.contrast_conversions import adapt_intensity_range
from stimupy.utils.utils import apply_bessel, round_to_vals

__all__ = [
    "sine",
    "square",
    "staircase",
    "bessel",
]


def resolve_grating_params(
    length=None,
    visual_angle=None,
    ppd=None,
    frequency=None,
    n_phases=None,
    phase_width=None,
    period="ignore",
    round_phase_width=True,
):
    """Resolve (if possible) spatial parameters for a grating

    Gratings take the regular resolution parameters (shape, ppd, visual_size).
    In addition, there has to be an additional specification
    of the frequency, number and with of phases (e.g., bars).
    This can be done in two ways:
    a phase_width (in degrees) and n_phases,
    and/or by specifying the spatial frequency of the grating (in cycles per degree)

    The total shape (in pixels) and visual size (in degrees) has to match the
    specification of the phases and their widths.
    Thus, not all 6 parameters have to be specified, as long as the both the resolution
    and the distribution of phases can be resolved.

    Note: all phases in a grating have the same width

    Parameters
    ----------
    length : Number, or None (default)
        length of grating, in pixels
    visual_angle :Number, or None (default)
        visual angle of grating, in degrees
    ppd : Number, or None (default)
        pixels per degree along the axis of grating
    frequency : Number, or None (default)
        frequency of grating, in cycles per degree
    n_phases : int, or None (default)
        number of phases (e.g., bars), i.e., half the number of full periods
    phase_width : Number, or None (default)
        extend of a single phase (e.g., bar), in degrees
    period : "even", "odd", "either", "ignore" (default)
        whether to ensure the grating has "even" number of phases,
        "odd" number of phases, or "either" even/odd,
        or no guarantees ("ignore")
    round_phase_width : bool,
        whether to round phase_width to an integer number of pixels, by default True


    Returns
    -------
    dict[str, Any]
        dictionary with all six resolution & size parameters resolved.
    """
    old_frequency = deepcopy(frequency)
    old_n_phases = deepcopy(n_phases)
    old_phase_width = deepcopy(phase_width)

    if period not in ["ignore", "even", "odd", "either"]:
        raise TypeError(f"period not understood: {period}")

    # Try to resolve resolution
    try:
        length, visual_angle, ppd = resolution.resolve_1D(
            length=length, visual_angle=visual_angle, ppd=ppd
        )
    except resolution.TooManyUnknownsError:
        ppd = ppd
        length = length
        visual_angle = visual_angle

    # Try to resolve number and width(s) of phases:
    # Logic here is that phase_width expresses "degrees per phase",
    # which we can convert to "phases_per_degree"
    # phase_width = degrees_per_phase = 1 / phases_per_degree = 1 / (2*frequency)
    if phase_width is not None:
        phases_pd = 1 / phase_width
        if frequency is not None and phases_pd != 2 * frequency:
            raise ValueError(f"phase_width {phase_width} and frequency {frequency} don't match")
    elif frequency is not None:
        phases_pd = 2 * frequency
    else:  # both are None:
        phases_pd = None

    # Logic here is that phase_width expresses "degrees per phase",
    # which we can invert to phases_per_degree, analogous to ppd:
    # n_phases = phases_per_degree * n_degrees
    # is analogous to
    # pix = ppd * n_degrees
    # Thus we can resolve the number and spacing of phases also as a resolution
    try:
        n_phases, min_angle, phases_pd = resolution.resolve_1D(
            length=n_phases,
            visual_angle=visual_angle,
            ppd=phases_pd,
            round=False,
        )
    except Exception as e:
        raise Exception("Could not resolve grating frequency, phase_width, n_phases") from e

    # Now resolve resolution
    visual_angle = min_angle if visual_angle is None else visual_angle
    length, visual_angle, ppd = resolution.resolve_1D(
        length=length, visual_angle=visual_angle, ppd=ppd
    )

    # Ensure n phases fit in length?
    if period != "ignore":
        n_phases = round_n_phases(n_phases=n_phases, length=length, period=period)

        # Calculate other params again:
        n_phases, min_angle, phases_pd = resolution.resolve_1D(
            length=n_phases,
            visual_angle=visual_angle,
            ppd=None,
            round=False,
        )

    # Ensure each phase consists of integer number of pixels?
    phase_width = 1 / phases_pd
    if round_phase_width:
        phase_width = np.round(phase_width * ppd) / ppd
        n_phases = length / np.round(phase_width * ppd)

    # Calculate frequency
    frequency = 1 / (2 * phase_width)

    # Check & warn if we changed some params
    if period == "ignore":
        if old_n_phases is not None and n_phases != old_n_phases:
            warnings.warn(
                f"Adjusted n_phases={old_n_phases} -> {n_phases} because of poor resolution"
            )

        if old_phase_width is not None and phase_width != old_phase_width:
            warnings.warn(
                f"Adjusted phase_width={old_phase_width} -> {phase_width} because of poor"
                " resolution"
            )

        if old_frequency is not None and frequency != old_frequency:
            warnings.warn(
                f"Adjusted frequency={old_frequency} -> {frequency} because of poor resolution"
            )
    else:
        if old_n_phases is not None and n_phases != old_n_phases:
            warnings.warn(
                f"Adjusted n_phases={old_n_phases} -> {n_phases} because of period={period}"
            )

        if old_phase_width is not None and phase_width != old_phase_width:
            warnings.warn(
                f"Adjusted phase_width={old_phase_width} -> {phase_width} because of"
                f" period={period}"
            )

        if old_frequency is not None and frequency != old_frequency:
            warnings.warn(
                f"Adjusted frequency={old_frequency} -> {frequency} because of period={period}"
            )

    # Check that frequency does not exceed Nyquist limit:
    if frequency > (ppd / 2):
        raise ValueError(
            f"Grating frequency ({frequency}) should not exceed Nyquist limit {ppd/2} (ppd/2)"
        )

    # Accumulate edges of phases (rounding to avoid accumulation of
    # floating point imprecisions)
    edges = [*itertools.accumulate(itertools.repeat(phase_width, int(np.ceil(n_phases))))]
    edges = np.round(np.array(edges), 8)
    edges = list(edges)

    return {
        "length": length,
        "visual_angle": visual_angle,
        "ppd": ppd,
        "frequency": frequency,
        "phase_width": phase_width,
        "n_phases": n_phases,
        "edges": edges,
        "period": period,
    }


def round_n_phases(n_phases, length, period="either"):
    """Round n_phases of grating to integer, even, or odd -- taking into account pixels

    Finds the nearest integer (optionally limited to even or odd) n_phases
    that length (in pixels) can be divided into.
    Note that this maybe be quite far away from the input.


    Parameters
    ----------
    n_phases : int
        number of phases (e.g., bars), i.e., half the number of full periods
    length : Number
        length of grating, in pixels
    period : "even", "odd", "either" (default)
        whether to ensure the grating has "even" number of phases,
        "odd" number of phases, or "either" even/odd

    Returns
    -------
    n_phases : int
        rounded n_phases
    """

    # n_phases has to integer-divide length
    possible_phase_pix = np.array(list(int_factorize(length)))
    possible_n_phases = length / possible_phase_pix

    if period == "even":
        # only look at possible_n_phases that are even
        possible_n_phases = possible_n_phases[possible_n_phases % 2 == 0]
    elif period == "odd":
        # only look at possible_n_phases that are odd
        possible_n_phases = possible_n_phases[possible_n_phases % 2 != 0]

    if len(possible_n_phases) == 0:
        raise ValueError(f"Cannot fit {period} number of phases into {length} px")

    closest = possible_n_phases[np.argmin(np.abs(possible_n_phases - n_phases))]

    return int(closest)


def sine(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_phases=None,
    phase_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0.0,
    intensities=(0.0, 1.0),
    origin="center",
    distance_metric=None,
    round_phase_width=False,
):
    """Draw a sine-wave grating given a certain distance_metric

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle.
        For `distance_metric="angular"`, this is used as cycles-per-image.
    n_phases : int, or None (default)
        number of phases in the grating
    phase_width : Number, or None (default)
        width of a single phase, in degrees visual angle
    period : "even", "odd", "either", "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees, by default 0.0
    intensities : Sequence[float, float]
        min and max intensity of sine-wave, by default (0.0, 1.0).
    origin : "corner", "mean", or "center" (default)
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    distance_metric : str or None
        if "horizontal", use distance from origin in x-direction,
        if "vertical", use distance from origin in x-direction;
        if "oblique", use combined and rotated distance from origin in x-y;
        if "radial", use radial distance from origin,
        if "angular", use angular distance from origin,
        if "rectilinear", use rectilinear/cityblock/Manhattan distance from origin
    round_phase_width : bool
        if True, round width of bars given resolution, by default False.

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each bar (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """
    distance_metrics = ["horizontal", "vertical", "oblique", "radial", "angular", "rectilinear"]
    if distance_metric not in distance_metrics:
        raise ValueError(f"distance_metric needs to be one of {distance_metrics}")

    lst = [visual_size, ppd, shape, frequency, n_phases, phase_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating()' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_phases', 'phase_width'"
        )

    # Try to resolve resolution
    try:
        shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    except ValueError:
        ppd = resolution.validate_ppd(ppd)
        shape = resolution.validate_shape(shape)
        visual_size = resolution.validate_visual_size(visual_size)

    alpha = [np.abs(np.cos(np.deg2rad(rotation))), np.abs(np.sin(np.deg2rad(rotation)))]

    if shape.width is not None:
        length = np.round(alpha[0] * shape.width + alpha[1] * shape.height)
    else:
        length = None

    if visual_size.width is not None:
        visual_angle = alpha[0] * visual_size.width + alpha[1] * visual_size.height
    else:
        visual_angle = None

    if ppd.horizontal is not None:
        ppd_1D = ppd.horizontal
    else:
        ppd_1D = None

    if rotation % 90 != 0 and round_phase_width:
        round_phase_width = False
        warnings.warn("Rounding phase width is turned off for oblique gratings")

    if rotation % 90 != 0 and period != "ignore":
        period = "ignore"
        warnings.warn("Period ignored for oblique gratings")

    # Resolve params
    if distance_metric == "radial" or distance_metric == "rectilinear":
        if n_phases is not None:
            n_phases = n_phases * 2

    if distance_metric == "angular":
        params = resolve_grating_params(
            visual_angle=360,
            n_phases=n_phases,
            phase_width=phase_width,
            ppd=1,
            frequency=frequency / 360 if frequency is not None else None,
            period=period,
            round_phase_width=round_phase_width,
        )
    else:
        params = resolve_grating_params(
            length=length,
            visual_angle=visual_angle,
            n_phases=n_phases,
            phase_width=phase_width,
            ppd=ppd_1D,
            frequency=frequency,
            period=period,
            round_phase_width=round_phase_width,
        )

    length = params["length"]
    ppd_1D = params["ppd"]
    visual_angle = params["visual_angle"]
    frequency = params["frequency"]
    phase_width = params["phase_width"]
    n_phases = params["n_phases"]

    if distance_metric == "radial" or distance_metric == "rectilinear":
        n_phases = n_phases / 2

    # Determine size/shape of whole image
    if None in shape:
        shape = [length * alpha[1], length * alpha[0]]
        if np.round(alpha[1], 5) == 0:
            shape[0] = shape[1]
        if np.round(alpha[0], 5) == 0:
            shape[1] = shape[0]

    if None in ppd:
        ppd = (ppd_1D, ppd_1D)

    if None in visual_size:
        visual_size = resolution.visual_size_from_shape_ppd(shape=shape, ppd=ppd)

    shape = resolution.validate_shape(shape)
    visual_size = resolution.validate_visual_size(visual_size)
    ppd = resolution.validate_ppd(ppd)

    # Set up coordinates
    base = image_base(
        shape=shape, visual_size=visual_size, ppd=ppd, rotation=rotation, origin=origin
    )
    distances = base[distance_metric]
    distances = np.round(distances, 6)

    # Shift distances minimally to ensure proper behavior
    if origin == "corner":
        distances = adapt_intensity_range(distances, 1e-03, distances.max() - 1e-03)
    else:
        distances = adapt_intensity_range(
            distances, distances.min() - 1e-05, distances.max() - 1e-05
        )

    if distance_metric == "angular":
        distances = adapt_intensity_range(
            distances, distances.min() - 1e-05, n_phases * phase_width - 1e-05
        )

    # Draw image
    img = np.sin(frequency * 2 * np.pi * distances + np.deg2rad(phase_shift))
    img = adapt_intensity_range(img, intensities[0], intensities[1])

    # Create mask
    dmax = max(distances.max(), -distances.min()) + (phase_width / 1)
    if (
        origin == "corner"
        or distance_metric == "radial"
        or distance_metric == "rectilinear"
        or distance_metric == "angular"
    ):
        edges = np.arange(distances.min() + (phase_width / 2), dmax, phase_width)

        if origin == "mean":
            edges -= distances.min()
    else:
        edges_pos = np.arange(0 + (phase_width / 2), dmax, phase_width) + 1e-4
        edges_neg = -np.arange(0 + (phase_width / 2), dmax, phase_width) - 1e-4
        edges = np.unique(np.append(edges_neg[::-1], edges_pos))

    edges -= (((phase_shift) % 360) / 180) * phase_width
    # edges = np.round(edges, 8)

    regions = round_to_vals(distances, edges)
    mask = np.zeros(shape=regions.shape)
    for idx, val in enumerate(np.unique(regions)):
        mask = np.where(regions == val, idx + 1, mask)

    # Package and output
    stim = {
        "img": img,
        "grating_mask": mask.astype(int),
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "frequency": frequency,
        "n_phases": n_phases,
        "phase_width": phase_width,
        "period": period,
        "rotation": rotation,
        "phase_shift": phase_shift,
        "round_phase_width": round_phase_width,
        "origin": origin,
        "distance_metric": distance_metric,
        "intensities": intensities,
    }
    return stim


def square(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_phases=None,
    phase_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0.0,
    intensities=(0.0, 1.0),
    origin="center",
    distance_metric=None,
    round_phase_width=False,
):
    """Draw a square-wave grating given a certain distance_metric

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_phases : int, or None (default)
        number of phases in the grating
    phase_width : Number, or None (default)
        width of a single phase, in degrees visual angle
    period : "even", "odd", "either", "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees, by default 0.0
    intensities : Sequence[float, float]
        min and max intensity of sine-wave, by default (0.0, 1.0).
    origin : "corner", "mean", or "center" (default)
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    distance_metric : str or None
        if "horizontal", use distance from origin in x-direction,
        if "vertical", use distance from origin in x-direction;
        if "oblique", use combined and rotated distance from origin in x-y;
        if "radial", use radial distance from origin,
        if "angular", use angular distance from origin,
        if "rectilinear", use rectilinear/cityblock/Manhattan distance from origin
    round_phase_width : bool
        if True, round width of bars given resolution, by default False

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each bar (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """

    stim = sine(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_phases,
        phase_width=phase_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric=distance_metric,
    )

    # Round sine-wave to create square wave
    stim["img"] = round_to_vals(stim["img"], intensities)
    return stim


def staircase(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_phases=None,
    phase_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0.0,
    origin="center",
    distance_metric=None,
    round_phase_width=False,
    intensities=(0.0, 1.0),
):
    """Draw a luminance staircase

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_phases : int, or None (default)
        number of phases in the grating
    phase_width : Number, or None (default)
        width of a single phase, in degrees visual angle
    period : "even", "odd", "either", "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees, by default 0.o
    origin : "corner", "mean", or "center" (default)
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    distance_metric : str or None
        if "horizontal", use distance from origin in x-direction,
        if "vertical", use distance from origin in x-direction;
        if "oblique", use combined and rotated distance from origin in x-y;
        if "radial", use radial distance from origin,
        if "angular", use angular distance from origin,
        if "rectilinear", use rectilinear/cityblock/Manhattan distance from origin
    round_phase_width : bool
        if True, round width of bars given resolution, by default False
    intensities : Sequence[float, ...]
        if len(intensities)==2, intensity range of staircase (default 0.0, 1.0);
        if len(intensities)>2, intensity value for each phase.
        Can specify as many intensities as n_phases.
        If fewer intensities are passed than n_phases, cycles through intensities.

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each phase (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """

    stim = square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_phases,
        phase_width=phase_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric=distance_metric,
    )

    if len(intensities) == 2:
        intensities = np.linspace(intensities[0], intensities[1], stim["grating_mask"].max())

    # Use grating_mask to draw staircase
    stim["img"] = draw_regions(mask=stim["grating_mask"], intensities=intensities)
    stim["intensity_phases"] = intensities
    return stim


def bessel(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    order=0,
    intensities=(1.0, 0.0),
    origin="mean",
):
    """Draw a Bessel stimulus, i.e. draw circular rings following an nth order
    Bessel function of a given frequency.

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of circular grating, in cycles per degree
    order : int
        n-th order Bessel function
    intensities : (float, float)
        intensity values of rings, first value indicating center intensity
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    ----------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        empty mask (key: "ring_mask"),
        and additional keys containing stimulus parameters
    """
    if frequency is None:
        raise ValueError("bessel() missing argument 'frequency' which is not 'None'")

    base = image_base(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=0.0,
        origin=origin,
    )

    arr = base["radial"] * frequency * 2 * np.pi
    img = apply_bessel(arr, order=order)
    img = (img - img.min()) / (img.max() - img.min())
    img = img * (intensities[0] - intensities[1]) + intensities[1]

    stim = {
        "img": img,
        "ring_mask": np.zeros(base["shape"]).astype(int),
        "visual_size": base["visual_size"],
        "ppd": base["ppd"],
        "shape": base["shape"],
        "order": order,
        "frequency": frequency,
        "intensities": intensities,
        "origin": origin,
    }
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {"visual_size": 15, "ppd": 30, "intensities": (0.0, 1.0)}
    default_params.update(kwargs)

    grating_params = {
        # "phase_width": 3.5,
        "n_phases": 8,
        "period": "ignore",
        "phase_shift": 0,
        "origin": "center",
        "round_phase_width": False,
    }

    # fmt: off
    stimuli = {
        "waves_sine_horizontal": sine(**default_params, **grating_params, distance_metric="horizontal"),
        "waves_sine_vertical": sine(**default_params, **grating_params, distance_metric="vertical"),
        "waves_sine_oblique": sine(**default_params, **grating_params, distance_metric="oblique", rotation=30),
        "waves_sine_radial": sine(**default_params, **grating_params, distance_metric="radial"),
        "waves_sine_angular": sine(**default_params, **grating_params, distance_metric="angular"),
        "waves_sine_rectilinear": sine(**default_params, **grating_params, distance_metric="rectilinear"),

        "waves_square_horizontal": square(**default_params, **grating_params, distance_metric="horizontal"),
        "waves_square_vertical": square(**default_params, **grating_params, distance_metric="vertical"),
        "waves_square_oblique": square(**default_params, **grating_params, distance_metric="oblique", rotation=30),
        "waves_square_radial": square(**default_params, **grating_params, distance_metric="radial"),
        "waves_square_angular": square(**default_params, **grating_params, distance_metric="angular"),
        "waves_square_rectilinear": square(**default_params, **grating_params, distance_metric="rectilinear"),

        "waves_staircase_horizontal": staircase(**default_params, **grating_params, distance_metric="horizontal"),
        "waves_staircase_vertical": staircase(**default_params, **grating_params, distance_metric="vertical"),
        "waves_staircase_oblique": staircase(**default_params, **grating_params, distance_metric="oblique", rotation=30),
        "waves_staircase_radial": staircase(**default_params, **grating_params, distance_metric="radial"),
        "waves_staircase_angular": staircase(**default_params, **grating_params, distance_metric="angular"),
        "waves_staircase_rectilinear": staircase(**default_params, **grating_params, distance_metric="rectilinear"),

        "waves_bessel": bessel(**default_params, frequency=0.5),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Tests whether validation of resolution specification works as expected

Tests the combined specification of shape (in pixels), visual_size (deg) and ppd.
If this makes sense, i.e. roughly, int(visual_size * ppd) == shape,
function should pass without output.
If it does not make sense, function should raise a ResolutionError.

The tests do this through parameterization:
pytest runs each test for each of the specified parameter sets.

For the valid-input tests, each parameter set specifies the (valid) input,
and value (in canonical form) that should be returned.
If the actual returned value matches this expected return value, the test passes.

For the invalid-input tests, each parameter set specifies the (invalid) input,
and the function should raise a ResolutionError for each of these.
"""
import pytest

from stimupy.utils import resolution


@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    (
        ((1024, 1024), (32, 32), (32, 32)),
        ((1024, 1024), (32, 16), (32, 64)),
        ((512, 1024), (16, 64), (32, 16)),
        ((500, 400), (20, 40), (25, 10)),
        (1024, (32, 32), (32, 32)),
        ((1024, 1024), 32, (32, 32)),
        ((1024, 1024), (32, 32), 32),
    ),
)
def test_valid_resolution(shape, visual_size, ppd):
    resolution.valid_resolution(shape=shape, visual_size=visual_size, ppd=ppd)


@pytest.mark.parametrize(
    "shape, visual_size, ppd",
    (
        ((1024, 1024), (32, 32), (16, 16)),
        ((1024, 1024), (32, 16), (32, 32)),
        ((512, 1024), (32, 32), (32, 32)),
        ((500, 400), (20, 20), (25, 10)),
        (1024, (32, 16), (32, 32)),
        ((1024, 512), 32, (32, 30)),
        ((1024, 512), (32, 32), 32),
        ((1024, None), (32, 32), (32, 32)),
        ((1024, 1024), (32, None), (32, 32)),
        ((1024, 1024), (32, 32), (None, 32)),
        ((1024, None), (32, 32), (32, 32)),
        ((1024, 1024), (32, None), (32, 32)),
        (None, (32, 32), (32, 32)),
        ((1024, 1024), None, (32, 32)),
        ((1024, 1024), (32, 32), None),
    ),
)
def test_invalid_resolution(shape, visual_size, ppd):
    with pytest.raises(resolution.ResolutionError):
        resolution.valid_resolution(shape=shape, visual_size=visual_size, ppd=ppd)



import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from stimupy.utils import resolution

__all__ = [
    "plot_stim",
    "plot_stimuli",
    "plot_comparison",
]


def plot_stim(
    stim,
    mask=False,
    stim_name="stim",
    ax=None,
    vmin=0,
    vmax=1,
    save=None,
    units="deg",
    origin="mean",
):
    """Plot a stimulus

    Plots the stimulus-array (key: "img") directly from stim dict.
    Optionally also plots mask.

    Parameters
    ----------
    stim : dict
        stimulus dict containing stimulus-array (key: "img")
    mask : bool or str, optional
        If True, plot mask on top of stimulus image (default: False).
        If string is provided, plot this key from stimulus dictionary as mask
    stim_name : str, optional
        Stimulus name used for plotting (default: "stim")
    ax : Axis object, optional
        If not None (default), plot in the specified Axis object
    vmin : float, optional
        Minimal intensity value for plotting. The default is 0.
    vmax : float, optional
        Minimal intensity value for plotting. The default is 1.
    save : None or str, optional
        If None (default), do not save the plot.
        If string is provided, save plot under this name.
    units : "px", "deg" (default), or str
        what units to put on the axes, by default degrees visual angle ("deg").
        If a str other than "deg"(/"degrees") or "px"(/"pix"/"pixels") is passed,
        it must be the key to a tuple in stim

    Returns
    -------
    ax : Axis object
        If ax was passed and plotting is None, returns updated Axis object.

    """
    single_plot = False
    if ax is None:
        ax = plt.gca()
        single_plot = True

    # Figure out what units need to go on axes
    if units in ["px", "pix", "pixels"]:
        extent = [0, stim["img"].shape[1], stim["img"].shape[0], 0]
    elif units in ["deg", "degrees"]:
        if "visual_size" in stim:
            x, y = resolution.visual_size_to_axes(
                stim["visual_size"], shape=stim["img"].shape, origin=origin
            )
            extent = [x.min(), x.max(), y.max(), y.min()]
        else:
            warnings.warn("no visual_size provided")
            extent = [0, stim["img"].shape[1], stim["img"].shape[0], 0]
    elif units in stim.keys():
        if len(stim[units]) == 2:
            # provided 2 values for units, so assume formatted like visual_size
            x, y = resolution.visual_size_to_axes(
                stim[units], shape=stim["img"].shape, origin=origin
            )
            extent = [x.min(), x.max(), y.min(), y.max()]
        elif len(stim[units]) == 4:
            # provided 4 values for units, so assume proper formatting
            extent = stim[units]
        else:
            raise ValueError("extent should either contain 2 or 4 values")
    else:
        warnings.warn("units does not exist in dict, using pixel-extent")
        extent = [0, stim["img"].shape[1], 0, stim["img"].shape[0]]

    if not mask:
        ax.imshow(stim["img"], cmap="gray", vmin=vmin, vmax=vmax, extent=extent)
    else:
        img = stim["img"]
        if isinstance(mask, str):
            mask_keys = [mask]
        else:
            mask_keys = [key for key in stim.keys() if key.endswith("mask")]

        if len(mask_keys) == 0:
            ax.imshow(stim["img"], cmap="gray", vmin=vmin, vmax=vmax, extent=extent)
        else:
            # If target_mask exists, use it.
            if "target_mask" in mask_keys:
                mask = stim["target_mask"]
            else:
                mask = stim[mask_keys[0]]

            if (mask is None) or (len(np.unique(mask)) == 1):
                warnings.warn("mask is None or empty- cannot plot mask")
                ax.imshow(stim["img"], cmap="gray", vmin=vmin, vmax=vmax, extent=extent)

            else:
                img = np.dstack([img, img, img])
                mask = np.dstack([mask, mask, mask])

                if np.unique(mask).size >= 20:
                    colormap = plt.cm.colors.ListedColormap(np.random.rand(mask.max() + 1, 3))
                elif np.unique(mask).size >= 10 and np.unique(mask).size < 20:
                    colormap = plt.cm.tab20
                else:
                    colormap = plt.cm.tab10

                for idx in np.unique(mask)[np.unique(mask) > 0]:
                    color = colormap.colors[idx]
                    color = np.reshape(color, (1, 1, 3))
                    img = np.where(mask == idx, color, img)
                ax.imshow(img, extent=extent)

                # Colorbar for mask indices
                bounds = list(np.unique(mask))
                norm = mpl.colors.BoundaryNorm(bounds, len(bounds) + 1, extend="both")
                plt.colorbar(
                    mpl.cm.ScalarMappable(norm=norm, cmap=colormap),
                    ax=ax,
                )

    ax.set_title(label=stim_name)

    if save is None and single_plot:
        plt.show()
        return ax
    elif save is None and not single_plot:
        return ax
    elif isinstance(save, str):
        plt.savefig(save)
        plt.close()
    else:
        raise ValueError("save can be None or a filepath")


def plot_stimuli(
    stims,
    mask=False,
    vmin=0,
    vmax=1,
    save=None,
    units="deg",
):
    """Plot multiple stimuli

    Plots the stimulus-arrays (keys: "img") directly from stim dicts.
    Arranges stimuli in a grid.
    Optionally also plots masks.

    Parameters
    ----------
    stims : dict of dicts
        dictionary composed of stimulus dicts containing stimulus-array (key: "img")
    mask : bool or str, optional
        If True, plot mask on top of stimulus image (default: False).
        If string is provided, plot this key from stimulus dictionary as mask
    vmin : float, optional
        Minimal intensity value for plotting. The default is 0.
    vmax : float, optional
        Minimal intensity value for plotting. The default is 1.
    save : None or str, optional
        If None (default), do not save the plot.
        If string is provided, save plot under this name.
    units : "px", "deg" (default), or str
        what units to put on the axes, by default degrees visual angle ("deg").
        If a str other than "deg"(/"degrees") or "px"(/"pix"/"pixels") is passed,
        it must be the key to a tuple in stim
    """

    # Plot each stimulus+mask
    n_stim = int(np.ceil(np.sqrt(len(stims))))
    F = plt.figure(figsize=(n_stim * 3, n_stim * 3))
    for i, (stim_name, stim) in enumerate(stims.items()):
        ax = F.add_subplot(n_stim, n_stim, i + 1)
        plot_stim(
            stim,
            mask,
            stim_name=stim_name,
            ax=ax,
            vmin=vmin,
            vmax=vmax,
            save=None,
            units=units,
        )

    plt.tight_layout()

    if save is None:
        plt.show()
    elif isinstance(save, str):
        plt.savefig(save)
        plt.close()
    else:
        raise ValueError("save can be None or a filepath")


def plot_comparison(original_img, new_img):
    """Plots visual comparison of two image-arrays

    Parameters
    ----------
    original_img : numpy.ndarray
        original image-array
    new_img : numpy.ndarray
        new image-array

    Returns
    -------
    matplotlib.Figure
        Figure containing plots of images, and their comparison(s)
    """
    vmin, vmax = 0, 1

    fig = plt.figure(figsize=(20, 6))
    plt.subplot(1, 4, 1)
    plt.imshow(original_img, cmap="gray", vmin=vmin, vmax=vmax)
    plt.title("Original")

    plt.subplot(1, 4, 2)
    plt.imshow(new_img, cmap="gray", vmin=vmin, vmax=vmax)
    plt.title("New")

    plt.subplot(1, 4, 3)
    plt.imshow(original_img - new_img, cmap="coolwarm", vmin=-vmax, vmax=vmax)
    plt.colorbar()
    num_pix_off = np.prod(original_img.shape) - np.sum(np.isclose(original_img, new_img))
    plt.title(f"Difference: {num_pix_off} pix")

    plt.subplot(1, 4, 4)
    plt.plot(original_img[:, 128], label="Original")
    plt.plot(new_img[:, 128], label="New")
    plt.legend()

    return fig



import copy

import numpy as np

from stimupy.components import lines
from stimupy.utils import make_two_sided, resolution

__all__ = [
    "mueller_lyer",
    "two_sided",
]


def mueller_lyer(
    visual_size=None,
    ppd=None,
    shape=None,
    outer_lines_length=None,
    outer_lines_angle=45,
    target_length=None,
    line_width=0,
    intensity_outer_lines=1.0,
    intensity_target=0.5,
    intensity_background=0.0,
):
    """Mueller-Lyer's (1896) illusion

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    outer_lines_length : Number
        length of outer lines in degrees visual angle
    outer_lines_angle : Number (optional)
        angle of outer lines in degrees, by default 45. Must be between -180 and 180 degrees.
    target_length : Number
        length of target line in degrees visual angle
    line_width : Number (optional)
        line width in degrees visual angle; if 0 (default), line width is 1 px
    intensity_outer_lines : Number (optional)
        intensity value of outer lines, by default 0.01
    intensity_target : Number (optional)
        intensity value of target line, by default 0.5
    intensity_background : Number (optional)
        intensity value of background, by default 0.0

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Mueller-Lyer, F. (1896).
        Zur Lehre von den optischen Taeuschungen.
        Ueber Kontrast und Konfluxion.
        Zeitschrift fuer Psychologie und Physiologie der Sinnesorgane, IX, 1-16.
    """
    if outer_lines_length is None:
        raise ValueError(
            "mueller_lyer() missing argument 'outer_lines_length' which is not 'None'"
        )
    if target_length is None:
        raise ValueError("mueller_lyer() missing argument 'target_length' which is not 'None'")

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    if outer_lines_angle > 180:
        outer_lines_angle -= 360

    if outer_lines_angle < -180 or outer_lines_angle > 180:
        raise ValueError("outer_lines_angle should be between -180 and 180 deg")

    angle1 = copy.deepcopy(outer_lines_angle) + 90
    angle2 = -angle1 - 180
    angle4 = copy.deepcopy(outer_lines_angle) - 90
    angle3 = -angle4 - 180

    target_line = lines.line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=None,
        line_length=target_length,
        line_width=line_width,
        rotation=90,
        intensity_line=intensity_target - intensity_background,
        intensity_background=0,
        origin="center",
    )

    oline1 = lines.line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=(0, -target_length / 2),
        line_length=outer_lines_length,
        line_width=line_width,
        rotation=angle1,
        intensity_line=intensity_outer_lines - intensity_background,
        intensity_background=0,
        origin="center",
    )

    oline2 = lines.line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=(0, -target_length / 2),
        line_length=outer_lines_length,
        line_width=line_width,
        rotation=angle2,
        intensity_line=intensity_outer_lines - intensity_background,
        intensity_background=0,
        origin="center",
    )

    oline3 = lines.line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=(0, target_length / 2),
        line_length=outer_lines_length,
        line_width=line_width,
        rotation=angle3,
        intensity_line=intensity_outer_lines - intensity_background,
        intensity_background=0,
        origin="center",
    )

    oline4 = lines.line(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        line_position=(0, target_length / 2),
        line_length=outer_lines_length,
        line_width=line_width,
        rotation=angle4,
        intensity_line=intensity_outer_lines - intensity_background,
        intensity_background=0,
        origin="center",
    )

    # Add outer lines together
    olines = oline1["img"] + oline2["img"] + oline3["img"] + oline4["img"]
    omasks1 = oline1["line_mask"] * 2 + oline2["line_mask"] * 3
    omasks2 = oline3["line_mask"] * 4 + oline4["line_mask"] * 5

    target_line["img"] += olines
    target_line["img"] = np.where(
        target_line["img"] > intensity_outer_lines, intensity_outer_lines, target_line["img"]
    )

    target_line["line_mask"] += omasks1
    target_line["line_mask"] = np.where(target_line["line_mask"] > 3, 3, target_line["line_mask"])
    target_line["line_mask"] += omasks2
    target_line["line_mask"] = np.where(target_line["line_mask"] > 5, 5, target_line["line_mask"])

    target_line["target_mask"] = np.where(target_line["line_mask"] == 1, 1, 0).astype(int)
    target_line["outer_lines_length"] = outer_lines_length
    target_line["outer_lines_angle"] = outer_lines_angle
    target_line["target_length"] = target_length
    target_line["line_width"] = line_width
    target_line["intensity_outer_lines"] = intensity_outer_lines
    target_line["intensity_target"] = intensity_target
    target_line["intensity_background"] = intensity_background
    return target_line


two_sided = make_two_sided(
    mueller_lyer,
    two_sided_params=(
        "outer_lines_length",
        "outer_lines_angle",
        "target_length",
        "line_width",
        "intensity_outer_lines",
        "intensity_target",
    ),
)


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    stim_params = {
        "outer_lines_length": 1,
        "target_length": 3,
        "line_width": 0.1,
    }

    # fmt: off
    stimuli = {
        "mueller_lyer": mueller_lyer(**default_params, **stim_params, outer_lines_angle=45),
        "mueller_lyer_2sided": two_sided(**default_params, **stim_params, outer_lines_angle=(30, 135)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import copy

import numpy as np

from .utils import resolution

__all__ = [
    "add_padding",
    "remove_padding",
    "pad_by_visual_size",
    "pad_to_visual_size",
    "pad_by_shape",
    "pad_to_shape",
    "pad_dict_by_visual_size",
    "pad_dict_to_visual_size",
    "pad_dict_by_shape",
    "pad_dict_to_shape",
]


def remove_padding(arr, c):
    """
    Remove padding by c

    Parameters
    ----------
    arr : numpy.ndarray
        input array
    c : int
        padding amount.

    Returns
    -------
    arr : numpy.ndarray
        reduced array

    """
    array_shape = arr.shape
    arr = arr[c : array_shape[0] - c, c : array_shape[1] - c]
    return arr


def add_padding(arr, c, val):
    """
    Remove padding by c

    Parameters
    ----------
    arr : numpy.ndarray
        input array
    c : int
        padding amount
    val : float
        background value

    Returns
    -------
    arr : numpy.ndarray
        padded array

    """
    h, w = arr.shape
    new_arr = np.ones([h + c * 2, w + c * 2]) * val
    new_arr[c : h + c, c : w + c] = arr
    return new_arr


def pad_by_visual_size(img, padding, ppd, pad_value=0.0):
    """Pad image by specified degrees of visual angle

    Can specify different amount (before, after) each axis.

    Parameters
    ----------
    img : numpy.ndarray
        image-array to be padded
    padding : float, or Sequence[float, float], or Sequence[Sequence[float, float], ...]
        amount of padding, in degrees visual angle, in each direction:
        ((before_1, after_1), … (before_N, after_N)) unique pad widths for each axis
        (float,) or float is a shortcut for before = after = pad width for all axes.
    ppd : Sequence[Number] or Sequence[Number, Number]
        pixels per degree
    pad_value : Numeric, optional
        value to pad with, by default 0.0

    Returns
    -------
    numpy.ndarray
        img padded by the specified amount(s)

    See also
    ---------
    stimupy.utils.resolution
    """

    # Broadcast to ((before_1, after_1),...(before_N, after_N))
    padding_degs = np.broadcast_to(padding, (img.ndim, 2))

    # ppd in canonical form
    ppd = resolution.validate_ppd(ppd)

    # Convert to shape in pixels
    padding_px = []
    for axis in padding_degs:
        shape = [
            resolution.length_from_visual_angle_ppd(i, ppd) if i > 0 else 0
            for i, ppd in zip(axis, ppd)
        ]
        padding_px.append(shape)

    # Pad by shape in pixels
    return pad_by_shape(img=img, padding=padding_px, pad_value=pad_value)


def pad_to_visual_size(img, visual_size, ppd, pad_value=0):
    """Pad image to specified visual size in degrees visual angle

    Parameters
    ----------
    img : numpy.ndarray
        image-array to be padded
    visual_size : Sequence[int, int, ...]
        desired visual size (in degrees visual angle) of img after padding
    ppd : Sequence[Number] or Sequence[Number, Number]
        pixels per degree
    pad_value : Numeric, optional
        value to pad with, by default 0.0

    Returns
    -------
    numpy.ndarray
        img padded by the specified amount(s)

    See also
    ---------
    stimupy.utils.resolution
    """

    # visual_size to shape
    shape = resolution.shape_from_visual_size_ppd(visual_size=visual_size, ppd=ppd)

    # pad
    return pad_to_shape(img=img, shape=shape, pad_value=pad_value)


def pad_by_shape(img, padding, pad_value=0):
    """Pad image by specified amount(s) of pixels

    Can specify different amount (before, after) each axis.

    Parameters
    ----------
    img : numpy.ndarray
        image-array to be padded
    padding : int, or Sequence[int, int], or Sequence[Sequence[int, int], ...]
        amount of padding, in pixels, in each direction:
        ((before_1, after_1), … (before_N, after_N)) unique pad widths for each axis
        (int,) or int is a shortcut for before = after = pad width for all axes.
    pad_val : float, optional
        value to pad with, by default 0.0

    Returns
    -------
    numpy.ndarray
        img padded by the specified amount(s)
    """

    # Ensure padding is in integers
    padding = np.array(padding, dtype=np.int32)

    return np.pad(img, padding, mode="constant", constant_values=pad_value)


def pad_to_shape(img, shape, pad_value=0):
    """Pad image to a resulting specified shape in pixels

    Parameters
    ----------
    img : numpy.ndarray
        image-array to be padded
    shape : Sequence[int, int, ...]
        desired shape of img after padding
    pad_value : float, optional
        value to pad with, by default 0.0

    Returns
    -------
    numpy.ndarray
        img padded to specified shape

    Raises
    ------
    ValueError
        if img.shape already exceeds shape
    """

    if np.any(img.shape > shape):
        raise ValueError("img is bigger than size after padding")

    padding_per_axis = np.array(shape) - np.array(img.shape)
    padding_before = padding_per_axis // 2
    padding_after = padding_per_axis - padding_before
    padding = np.stack([padding_before, padding_after]).T

    return pad_by_shape(
        img,
        padding=padding,
        pad_value=pad_value,
    )


# %%
#################################
#         Dictionaries          #
#################################
def pad_dict_by_visual_size(dct, padding, ppd, pad_value=0.0, keys=("img", "*mask")):
    """Pad images in dictionary by specified degrees of visual angle

    Can specify different amount (before, after) each axis.

    Parameters
    ----------
    dct : dict
        dict containing image-arrays to be padded
    padding : float, or Sequence[float, float], or Sequence[Sequence[float, float], ...]
        amount of padding, in degrees visual angle, in each direction:
        ((before_1, after_1), … (before_N, after_N)) unique pad widths for each axis
        (float,) or float is a shortcut for before = after = pad width for all axes.
    ppd : Sequence[Number] or Sequence[Number, Number]
        pixels per degree
    pad_value : Numeric, optional
        value to pad with, by default 0.0
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with larger key-arrays and updated keys for
        "visual_size" and "shape"
    """
    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)

    if isinstance(keys, str):
        keys = (keys,)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                # Add mask which indicates padded region
                new_dict["pad_mask"] = pad_by_visual_size(
                    np.zeros(img.shape), padding, ppd, 1
                ).astype(int)

                if key.endswith("mask"):
                    img = pad_by_visual_size(img, padding, ppd, 0)
                    img = img.astype(int)
                else:
                    img = pad_by_visual_size(img, padding, ppd, pad_value)
                new_dict[key] = img

    # Update resolution
    new_dict["visual_size"] = resolution.visual_size_from_shape_ppd(img.shape, ppd)
    new_dict["shape"] = resolution.validate_shape(img.shape)
    return new_dict


def pad_dict_to_visual_size(dct, visual_size, ppd, pad_value=0, keys=("img", "*mask")):
    """Pad images in dictionary to specified visual size in degrees visual angle

    Parameters
    ----------
    dct : dict
        dict containing image-arrays to be padded
    visual_size : Sequence[int, int, ...]
        desired visual size (in degrees visual angle) of img after padding
    ppd : Sequence[Number] or Sequence[Number, Number]
        pixels per degree
    pad_value : Numeric, optional
        value to pad with, by default 0.0
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with larger key-arrays and updated keys for
        "visual_size" and "shape"
    """

    # visual_size to shape
    shape = resolution.shape_from_visual_size_ppd(visual_size=visual_size, ppd=ppd)

    # pad
    padded_dict = pad_dict_to_shape(dct=dct, shape=shape, pad_value=pad_value, keys=keys)

    # update resolution
    padded_dict["ppd"] = ppd
    padded_dict["visual_size"] = visual_size

    return padded_dict


def pad_dict_by_shape(dct, padding, pad_value=0, keys=("img", "*mask")):
    """Pad images in dictionary by specified amount(s) of pixels
    Can specify different amount (before, after) each axis.

    Parameters
    ----------
    dct : dict
        dict containing image-arrays to be padded
    padding : int, or Sequence[int, int], or Sequence[Sequence[int, int], ...]
        amount of padding, in pixels, in each direction:
        ((before_1, after_1), … (before_N, after_N)) unique pad widths for each axis
        (int,) or int is a shortcut for before = after = pad width for all axes.
    pad_val : float, optional
        value to pad with, by default 0.0
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with larger key-arrays and updated keys for
        "visual_size" and "shape"
    """
    # Ensure padding is in integers
    padding = np.array(padding, dtype=np.int32)

    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)

    if isinstance(keys, str):
        keys = (keys,)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                # Add mask which indicates padded region
                new_dict["pad_mask"] = np.pad(
                    np.zeros(img.shape), padding, mode="constant", constant_values=1
                ).astype(int)

                if key.endswith("mask"):
                    img = np.pad(img, padding, mode="constant", constant_values=0)
                    img = img.astype(int)
                else:
                    img = np.pad(img, padding, mode="constant", constant_values=pad_value)
                new_dict[key] = img

    # Update visual_size and shape-keys
    new_dict["shape"] = resolution.validate_shape(img.shape)
    if "ppd" in dct.keys():
        new_dict["visual_size"] = resolution.visual_size_from_shape_ppd(img.shape, dct["ppd"])
    return new_dict


def pad_dict_to_shape(dct, shape, pad_value=0, keys=("img", "*mask")):
    """Pad images in dictionary to a resulting specified shape in pixels

    Parameters
    ----------
    dct : dict
        dict containing image-arrays to be padded
    shape : Sequence[int, int, ...]
        desired shape of img after padding
    pad_value : float, optional
        value to pad with, by default 0.0
    keys : Sequence[String, String] or String
        keys in dict for images to be padded

    Returns
    -------
    dict[str, Any]
        same as input dict but with larger key-arrays and updated keys for
        "visual_size" and "shape"

    Raises
    ------
    ValueError
        if img.shape already exceeds shape
    """
    # Create deepcopy to not override existing dict
    new_dict = copy.deepcopy(dct)

    if isinstance(keys, str):
        keys = (keys,)

    # Find relevant keys
    keys = [
        dkey
        for key in keys
        for dkey in dct.keys()
        if ((dkey == key) or ((dkey.endswith(key[1::])) and (key.startswith("*"))))
    ]

    for key in dct.keys():
        if key in keys:
            img = dct[key]
            if isinstance(img, np.ndarray):
                if np.any(img.shape > shape):
                    raise ValueError("img is bigger than size after padding")

                padding_per_axis = np.array(shape) - np.array(img.shape)
                padding_before = padding_per_axis // 2
                padding_after = padding_per_axis - padding_before
                padding = np.stack([padding_before, padding_after]).T

                # Add mask which indicates padded region
                new_dict["pad_mask"] = pad_by_shape(
                    np.zeros(img.shape), padding=padding, pad_value=1
                ).astype(int)

                if key.endswith("mask"):
                    img = pad_by_shape(img, padding=padding, pad_value=0)
                    img = img.astype(int)
                else:
                    img = pad_by_shape(img, padding=padding, pad_value=pad_value)
                new_dict[key] = img

    # Update resolution
    new_dict["shape"] = resolution.validate_shape(shape)
    if "ppd" in dct.keys():
        new_dict["visual_size"] = resolution.visual_size_from_shape_ppd(shape, dct["ppd"])
    return new_dict



from stimupy.components.frames import frames
from stimupy.components.radials import rings
from stimupy.stimuli import place_targets
from stimupy.stimuli.waves import square_radial as circular
from stimupy.stimuli.waves import square_rectilinear as rectangular
from stimupy.utils import make_two_sided

__all__ = [
    "circular",
    "circular_generalized",
    "circular_two_sided",
    "rectangular",
    "rectangular_generalized",
    "rectangular_two_sided",
]


def circular_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    intensity_rings=(0.0, 1.0),
    intensity_background=0.5,
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
):
    """Sequential set of circular rings with specified radii and targets

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number] or None (default)
        radii of each ring, in degrees visual angle
    intensity_rings : Sequence[float, float]
        intensities of rings, by default (1.0, 0.0)
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    target_indices : int, or Sequence[int, ...]
        indices rings where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each ring (key: "target_mask"),
        and additional keys containing stimulus parameters
    """

    # Rings component
    stim = rings(
        radii=radii,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        intensity_rings=intensity_rings,
        intensity_background=intensity_background,
        origin=origin,
    )

    # Place target(s)
    stim = place_targets(
        stim=stim,
        element_mask_key="ring_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )

    return stim


circular_two_sided = make_two_sided(
    circular, two_sided_params=("intensity_rings", "intensity_target", "target_indices")
)


def rectangular_generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    radii=None,
    intensity_frames=(0.0, 1.0),
    intensity_background=0.5,
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    rotation=0.0,
):
    """Draw sequential set of square frames with specified radii and targets

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    radii : Sequence[Number] or None (default)
        radii of each frame, in degrees visual angle
    intensity_frames : Sequence[float, float]
        min and max intensity of square-wave, by default (0.0, 1.0)
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    target_indices : int, or Sequence[int, ...]
        indices frames where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)


    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each frame (key: "target_mask"),
        and additional keys containing stimulus parameters
    """

    # Frames component
    stim = frames(
        radii=radii,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        intensity_frames=intensity_frames,
        intensity_background=intensity_background,
        origin=origin,
        rotation=rotation,
    )
    stim["intensity_target"] = intensity_target

    # Place target(s)
    stim = place_targets(
        stim=stim,
        element_mask_key="frame_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )

    return stim


rectangular_two_sided = make_two_sided(
    rectangular, two_sided_params=("intensity_frames", "intensity_target", "target_indices")
)


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "rings_circular": circular(**default_params, frequency=1.0),
        "rings_circular_with_targets": circular(**default_params, frequency=1.0, target_indices=3),
        "rings_circular_clipped": circular(**default_params, frequency=1.0, clip=True),
        "rings_circular_2sided": circular_two_sided(**default_params, frequency=1.0, intensity_rings=((0.0, 1.0), (1.0, 0.0)), clip=True, target_indices=2),

        "rings_rectangular": rectangular(**default_params, frequency=1.0),
        "rings_rectangular_with_targets": rectangular(**default_params, frequency=1.0, target_indices=3),
        "rings_rectangular_clipped": rectangular(**default_params, frequency=1.0, clip=True),
        "rings_rectangular_2sided": rectangular_two_sided(**default_params, frequency=1.0, intensity_frames=((0.0, 1.0), (1.0, 0.0)), clip=True, target_indices=2),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Stimuli from Domijan (2015) https://doi.org/10.3389/fnhum.2015.00093

This module reproduces all of the stimuli used by Domijan (2015)
as they were provided to the model described in that paper.
Since the stimulus sizes were only defined in pixel-space,
there is some ambiguity with respect to the stimulus sizes in
degrees visual angle.
To help solve this ambiguity, we approximated a realistic resolution
of the stimuli (ppd = 10) which is set as default value.
However, because of the ambiguity, it is possible to change the
stimulus sizes by providing at least two of the following: a shape
(in pixels), a visual_size (in degrees) and/or a resolution (in ppd).

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.domijan2015.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/domijan2015.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.domijan2015 import *

References
-----------
Domijan, D. (2015).
    A neurocomputational account
    of the role of contour facilitation in brightness perception.
    Frontiers in Human Neuroscience, 9, 93.
    https://doi.org/10.3389/fnhum.2015.00093
"""

import numpy as np

import stimupy
from stimupy.utils import pad_dict_by_visual_size, pad_dict_to_shape, resolution, stack_dicts

__all__ = [
    "dungeon",
    "cube",
    "grating",
    "rings",
    "bullseye",
    "simultaneous_brightness_contrast",
    "white",
    "benary",
    "todorovic",
    "checkerboard_contrast_contrast",
    "checkerboard",
    "checkerboard_extended",
    "white_yazdanbakhsh",
    "white_anderson",
    "white_howe",
]

# Default values:
PPD = 10
PAD = True
SHAPES = {
    "dungeon": (110, 220),
    "cube": (100, 200),
    "grating": (100, 220),
    "rings": (100, 200),
    "bullseye": (100, 200),
    "simultaneous_brightness_contrast": (100, 200),
    "white": (80, 80),
    "benary": (100, 100),
    "todorovic": (100, 200),
    "checkerboard_contrast_contrast": (80, 160),
    "checkerboard": (80, 80),
    "checkerboard_extended": (80, 80),
    "white_yazdanbakhsh": (80, 80),
    "white_anderson": (100, 100),
    "white_howe": (100, 100),
}

VSIZES = {
    "dungeon": resolution.visual_size_from_shape_ppd(shape=SHAPES["dungeon"], ppd=PPD),
    "cube": resolution.visual_size_from_shape_ppd(shape=SHAPES["cube"], ppd=PPD),
    "grating": resolution.visual_size_from_shape_ppd(shape=SHAPES["grating"], ppd=PPD),
    "rings": resolution.visual_size_from_shape_ppd(shape=SHAPES["rings"], ppd=PPD),
    "bullseye": resolution.visual_size_from_shape_ppd(shape=SHAPES["bullseye"], ppd=PPD),
    "simultaneous_brightness_contrast": resolution.visual_size_from_shape_ppd(
        shape=SHAPES["simultaneous_brightness_contrast"], ppd=PPD
    ),
    "white": resolution.visual_size_from_shape_ppd(shape=SHAPES["white"], ppd=PPD),
    "benary": resolution.visual_size_from_shape_ppd(shape=SHAPES["benary"], ppd=PPD),
    "todorovic": resolution.visual_size_from_shape_ppd(shape=SHAPES["todorovic"], ppd=PPD),
    "checkerboard_contrast_contrast": resolution.visual_size_from_shape_ppd(
        shape=SHAPES["checkerboard_contrast_contrast"], ppd=PPD
    ),
    "checkerboard": resolution.visual_size_from_shape_ppd(shape=SHAPES["checkerboard"], ppd=PPD),
    "checkerboard_extended": resolution.visual_size_from_shape_ppd(
        shape=SHAPES["checkerboard_extended"], ppd=PPD
    ),
    "white_yazdanbakhsh": resolution.visual_size_from_shape_ppd(
        shape=SHAPES["white_yazdanbakhsh"], ppd=PPD
    ),
    "white_anderson": resolution.visual_size_from_shape_ppd(
        shape=SHAPES["white_anderson"], ppd=PPD
    ),
    "white_howe": resolution.visual_size_from_shape_ppd(shape=SHAPES["white_howe"], ppd=PPD),
}

v1, v2, v3 = 0.0, 0.5, 1.0


def gen_all(ppd=PPD, skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating domijan2015.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func(ppd=ppd)

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def resolve(shape, visual_size, ppd, original_visual_size):
    # Put in canonical form
    shape = resolution.validate_shape(shape)
    visual_size = resolution.validate_visual_size(visual_size)
    ppd = resolution.validate_ppd(ppd)

    # Try to resolve height; get resizing_factor from that
    try:
        _, visual_angle, ppd1 = resolution.resolve_1D(
            length=shape.height, visual_angle=visual_size.height, ppd=ppd.vertical
        )
        v1 = visual_angle / original_visual_size[0]
    except Exception:
        v1 = None
        ppd1 = None

    # Try to resolve width; get resizing_factor from that
    try:
        _, visual_angle, ppd2 = resolution.resolve_1D(
            length=shape.width, visual_angle=visual_size.width, ppd=ppd.horizontal
        )
        v2 = visual_angle / original_visual_size[1]
    except Exception:
        v2 = None
        ppd2 = None

    # Same resizing factor?
    visual_resize = [i for i in (v1, v2) if i is not None]
    visual_resize = np.unique(visual_resize)
    if len(visual_resize) != 1:
        # Different resizing factors -> not allowed
        raise ValueError(
            "Requested shape/visual_size is impossible given the stimulus defaults. "
            "Consider setting either the height or width to None"
        )
    else:
        # Same factor, resolve resolution using that
        visual_resize = visual_resize[0]
        ppd = np.unique([i for i in (ppd1, ppd2) if i is not None])
        shape, visual_size, ppd = resolution.resolve(
            shape, np.array(original_visual_size) * visual_resize, ppd
        )

    if ppd[0] % 2 != 0:
        raise ValueError("use an even ppd for domijan2015-stimuli")

    return shape, visual_size, ppd, visual_resize


def dungeon(visual_size=VSIZES["dungeon"], ppd=PPD, shape=SHAPES["dungeon"]):
    """Dungeon illusion, Domijan (2015) Fig 6A

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (11, 22)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (110, 220)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["dungeon"])
    ppd = ppd[0]

    # Define parameters for each side
    params = {
        "ppd": ppd,
        "n_cells": 5,
        "target_radius": 1,
        "cell_size": 1.0 * visual_resize,
    }

    # Generate each side
    stim1 = stimupy.dungeons.dungeon(
        **params,
        intensity_background=v1,
        intensity_grid=v3,
        intensity_target=v2,
    )
    stim2 = stimupy.dungeons.dungeon(
        **params,
        intensity_background=v3,
        intensity_grid=v1,
        intensity_target=v2,
    )

    # Pad and stack
    padding = np.array((0.9, 1.1)) * visual_resize
    stim1 = pad_dict_by_visual_size(stim1, padding, ppd, v1)
    stim2 = pad_dict_by_visual_size(stim2, padding, ppd, v3)
    stim = stack_dicts(stim1, stim2)

    params.update(
        original_shape=SHAPES["dungeon"],
        original_ppd=PPD,
        original_visual_size=VSIZES["dungeon"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def cube(visual_size=VSIZES["cube"], ppd=PPD, shape=SHAPES["cube"]):
    """Cube illusion, Domijan (2015) Fig 6B

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 20)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 200)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["cube"])
    ppd = ppd[0]

    params = {
        "ppd": ppd,
        "cell_thickness": 1.1 * visual_resize,
        "cell_lengths": np.array([1.8, 1.5, 1.5, 1.8]) * visual_resize,
        "target_indices": (1, 2),
        "cell_spacing": 0.5 * visual_resize,
    }

    stim1 = stimupy.cubes.varying_cells(
        **params,
        intensity_background=v1,
        intensity_cells=v3,
        intensity_target=v2,
    )
    stim2 = stimupy.cubes.varying_cells(
        **params,
        intensity_background=v3,
        intensity_cells=v1,
        intensity_target=v2,
    )

    # Pad and stack
    padding = np.array((0.9, 1.0)) * visual_resize
    stim1 = pad_dict_by_visual_size(stim1, padding, ppd, v1)
    stim2 = pad_dict_by_visual_size(stim2, padding, ppd, v3)
    stim = stack_dicts(stim1, stim2)

    params.update(
        original_shape=SHAPES["cube"],
        original_ppd=PPD,
        original_visual_size=VSIZES["cube"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def grating(visual_size=VSIZES["grating"], ppd=PPD, shape=SHAPES["grating"]):
    """Grating illusion, Domijan (2015) Fig 6C

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 22)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 220)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["grating"])
    visual_size_without_padding = (
        visual_size.height - ((0.9 + 1.0) * visual_resize),
        visual_size.width - (2 * (0.9 + 1.1) * visual_resize),
    )
    single_vissize = (visual_size_without_padding[0], visual_size_without_padding[1] / 2)

    params = {
        "visual_size": single_vissize,
        "ppd": ppd,
        "n_bars": 9,
        "target_indices": (5,),
        "bar_width": 1.0 * visual_resize,
    }

    stim1 = stimupy.waves.square_linear(
        **params,
        intensity_bars=(v1, v3),
        intensity_target=v2,
    )
    stim2 = stimupy.waves.square_linear(
        **params,
        intensity_bars=(v3, v1),
        intensity_target=v2,
    )

    # Pad and stack
    padding = np.array(((0.9, 1.0), (0.9, 1.1))) * visual_resize
    stim1 = pad_dict_by_visual_size(stim1, padding, ppd, v1)
    stim2 = pad_dict_by_visual_size(stim2, padding, ppd, v3)
    stim = stack_dicts(stim1, stim2)

    params.update(
        original_shape=SHAPES["grating"],
        original_ppd=PPD,
        original_visual_size=VSIZES["grating"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def rings(visual_size=VSIZES["rings"], ppd=PPD, shape=SHAPES["rings"]):
    """Ring patterns, Domijan (2015) Fig 7A

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 20)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 200)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["rings"])
    radii = np.array((0.55, 1.05, 1.55, 2.05, 2.55, 3.05, 3.55, 4.05)) * visual_resize

    params = {
        "visual_size": radii.max() * 2,
        "ppd": ppd,
        "radii": radii,
        "origin": "mean",
    }

    stim1 = stimupy.rings.rectangular_generalized(
        **params,
        target_indices=5,
        intensity_frames=(v1, v3),
        intensity_target=v2,
    )
    stim2 = stimupy.rings.rectangular_generalized(
        **params,
        target_indices=4,
        intensity_frames=(v1, v3),
        intensity_target=v2,
    )

    # Pad and stack
    stim1 = pad_dict_to_shape(stim1, shape=np.array(shape) / (1, 2), pad_value=v1)
    stim2 = pad_dict_to_shape(stim2, shape=np.array(shape) / (1, 2), pad_value=v1)
    stim = stack_dicts(stim1, stim2)

    params.update(
        original_shape=SHAPES["rings"],
        original_ppd=PPD,
        original_visual_size=VSIZES["rings"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def bullseye(visual_size=VSIZES["bullseye"], ppd=PPD, shape=SHAPES["bullseye"]):
    """Bullseye illusion, Domijan (2015) Fig 7B

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 20)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 200)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["bullseye"])
    radii = np.array((0.55, 1.05, 1.55, 2.05, 2.55, 3.05, 3.55, 4.05)) * visual_resize

    params = {
        "visual_size": radii.max() * 2,
        "ppd": ppd,
        "radii": radii,
        "origin": "mean",
    }

    stim1 = stimupy.bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(v1, v3),
        intensity_target=v2,
    )
    stim2 = stimupy.bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(v3, v1),
        intensity_target=v2,
    )

    # Pad and stack
    stim1 = pad_dict_to_shape(stim1, shape=np.array(shape) / (1, 2), pad_value=v1)
    stim2 = pad_dict_to_shape(stim2, shape=np.array(shape) / (1, 2), pad_value=v1)
    stim = stack_dicts(stim1, stim2)

    params.update(
        original_shape=SHAPES["bullseye"],
        original_ppd=PPD,
        original_visual_size=VSIZES["bullseye"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def simultaneous_brightness_contrast(
    visual_size=VSIZES["simultaneous_brightness_contrast"],
    ppd=PPD,
    shape=SHAPES["simultaneous_brightness_contrast"],
):
    """Simultaneous brightness contrast, Domijan (2015) Fig 7C

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 20)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 200)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(
        shape, visual_size, ppd, VSIZES["simultaneous_brightness_contrast"]
    )

    stim = stimupy.sbcs.generalized_two_sided(
        visual_size=visual_size,
        ppd=ppd,
        target_size=(2.1 * visual_resize, 2.1 * visual_resize),
        target_position=(3.8 * visual_resize, 3.8 * visual_resize),
        intensity_background=(v3, v1),
        intensity_target=v2,
    )

    stim.update(
        original_shape=SHAPES["simultaneous_brightness_contrast"],
        original_ppd=PPD,
        original_visual_size=VSIZES["simultaneous_brightness_contrast"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return stim


def white(visual_size=VSIZES["white"], ppd=PPD, pad=PAD, shape=SHAPES["white"]):
    """White stimulus, Domijan (2015) Fig 8A

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (8, 8)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (80, 80)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["white"])
    ppd = ppd[0]

    params = {
        "visual_size": visual_size,
        "ppd": ppd,
        "frequency": 4.0 / visual_size[1],
        "target_indices": (3, 6),
        "target_heights": 2.1 * visual_resize,
        "period": "even",
    }

    stim = stimupy.whites.white(
        **params,
        intensity_bars=(v3, v1),
        intensity_target=v2,
    )

    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim = pad_dict_by_visual_size(stim, padding, ppd, pad_value=v2)
        params["padding"] = padding

    original_shape_np = SHAPES["white"]
    original_shape = np.array(original_shape_np) + 20
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["white"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def benary(visual_size=VSIZES["benary"], ppd=PPD, shape=SHAPES["benary"]):
    """Benarys cross, Domijan (2015) Fig 8B

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 10)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 100)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["benary"])
    ppd = ppd[0]

    params = {
        "visual_size": 8.1 * visual_resize,
        "ppd": ppd,
        "cross_thickness": 2.1 * visual_resize,
        "target_size": 1.1 * visual_resize,
    }

    stim = stimupy.benarys.cross_rectangles(
        **params,
        intensity_background=v3,
        intensity_cross=v1,
        intensity_target=v2,
    )

    # Padding
    padding = np.array((0.9, 1.0)) * visual_resize
    stim = pad_dict_by_visual_size(stim, padding, ppd, pad_value=v3)

    params.update(
        original_shape=SHAPES["benary"],
        original_ppd=PPD,
        original_visual_size=VSIZES["benary"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def todorovic(visual_size=VSIZES["todorovic"], ppd=PPD, shape=SHAPES["todorovic"]):
    """Todorovic stimulus, Domijan (2015) Fig 9A

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 20)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 200)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Note: Compared to original, targets are moved by one pixel
    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["todorovic"])
    ppd = ppd[0]

    params = {
        "visual_size": visual_size[0],
        "ppd": ppd,
        "target_size": 4.1 * visual_resize,
        "target_position": 2.8 * visual_resize,
        "covers_size": 3.1 * visual_resize,
        "covers_offset": 2.0 * visual_resize,
    }

    stim1 = stimupy.todorovics.rectangle(
        **params,
        intensity_background=0.0,
        intensity_target=0.5,
        intensity_covers=1.0,
    )
    stim2 = stimupy.todorovics.rectangle(
        **params,
        intensity_background=1.0,
        intensity_target=0.5,
        intensity_covers=0.0,
    )

    # Stacking
    stim = stack_dicts(stim1, stim2)

    params.update(
        original_shape=SHAPES["todorovic"],
        original_ppd=PPD,
        original_visual_size=VSIZES["todorovic"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def checkerboard_contrast_contrast(
    visual_size=VSIZES["checkerboard_contrast_contrast"],
    ppd=PPD,
    shape=SHAPES["checkerboard_contrast_contrast"],
    pad=PAD,
):
    """Checkerboard contrast-contrast effect, Domijan (2015) Fig 9B

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (8, 16)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (80, 160)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(
        shape, visual_size, ppd, VSIZES["checkerboard_contrast_contrast"]
    )

    params = {
        "ppd": ppd,
        "check_visual_size": 1.0 * visual_resize,
        "target_shape": (4, 4),
        "tau": 0.5,
        "alpha": 0.5,
        "intensity_checks": (v1, v3),
    }

    # Large checkerboard, embedded target region
    stim1 = stimupy.checkerboards.contrast_contrast(
        **params,
        board_shape=(8, 8),
    )

    # Isolated target region (smaller checkerboard)
    stim2 = stimupy.checkerboards.contrast_contrast(
        **params,
        board_shape=(4, 4),
    )

    # Put smaller checkerboard on background (equally large as large checkerboard)
    stim2 = pad_dict_to_shape(stim2, stim1["img"].shape, pad_value=v2)

    # Overall padding
    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim1 = pad_dict_by_visual_size(stim1, padding, ppd=ppd, pad_value=v2)
        stim2 = pad_dict_by_visual_size(stim2, padding, ppd=ppd, pad_value=v2)
        params["padding"] = padding

    # Stacking
    stim = stack_dicts(stim1, stim2)

    # Output
    original_shape_np = np.array(SHAPES["checkerboard_contrast_contrast"])
    original_shape = original_shape_np + np.array((20, 40))
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["checkerboard_contrast_contrast"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
        board_shape_left=(8, 8),
        board_shape_right=(4, 4),
    )
    return {**stim, **params}


def checkerboard(
    visual_size=VSIZES["checkerboard"], ppd=PPD, shape=SHAPES["checkerboard"], pad=PAD
):
    """Classic checkerboard contrast with single-check targets, Domijan (2015) Fig 10A

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (8, 8)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (80, 80)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(
        shape, visual_size, ppd, VSIZES["checkerboard"]
    )

    params = {
        "ppd": ppd,
        "board_shape": (8, 8),
        "check_visual_size": (1.0 * visual_resize, 1.0 * visual_resize),
        "target_indices": [(3, 2), (5, 5)],
        "extend_targets": False,
        "intensity_checks": (v1, v3),
        "intensity_target": v2,
    }
    stim = stimupy.checkerboards.checkerboard(**params)

    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim = pad_dict_by_visual_size(stim, padding, ppd=ppd, pad_value=v2)
        params["padding"] = padding

    original_shape_np = SHAPES["checkerboard"]
    original_shape = np.array(original_shape_np) + 20
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["checkerboard"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )

    return {**stim, **params}


def checkerboard_extended(
    visual_size=VSIZES["checkerboard_extended"],
    ppd=PPD,
    shape=SHAPES["checkerboard_extended"],
    pad=PAD,
):
    """Checkerboard contrast with cross-like targets, Domijan (2015) Fig 10B

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (8, 8)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (80, 80)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(
        shape, visual_size, ppd, VSIZES["checkerboard_extended"]
    )

    params = {
        "ppd": ppd,
        "board_shape": (8, 8),
        "check_visual_size": (1.0 * visual_resize, 1.0 * visual_resize),
        "target_indices": [(3, 2), (5, 5)],
        "extend_targets": True,
        "intensity_checks": (v1, v3),
        "intensity_target": v2,
    }
    stim = stimupy.checkerboards.checkerboard(**params)

    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim = pad_dict_by_visual_size(stim, padding, ppd=ppd, pad_value=v2)
        params["padding"] = padding

    original_shape_np = SHAPES["checkerboard_extended"]
    original_shape = np.array(original_shape_np) + 20
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["checkerboard_extended"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )

    return {**stim, **params}


def white_yazdanbakhsh(
    visual_size=VSIZES["white_yazdanbakhsh"], ppd=PPD, shape=SHAPES["white_yazdanbakhsh"], pad=PAD
):
    """Yazdanbakhsh variation of White stimulus, Domijan (2015) Fig 11A

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (8, 8)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (80, 80)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    Yazdanbakhsh, A., Arabzadeh, E., Babadi, B., and Fazl, A. (2002).
        Munker-White-like illusions without T-junctions.
        Perception 31, 711-715.
        https://doi.org/10.1068/p3348
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(
        shape, visual_size, ppd, VSIZES["white_yazdanbakhsh"]
    )
    ppd = ppd[0]

    params = {
        "visual_size": visual_size,
        "ppd": ppd,
        "frequency": 4.0 / visual_size[1],
        "target_indices_top": (3,),
        "target_indices_bottom": (6,),
        "target_center_offset": 0.0,
        "target_heights": visual_size[0] / 4.0,
        "gap_size": visual_size[0] / 10.0,
        "period": "even",
    }

    stim = stimupy.whites.yazdanbakhsh(
        **params,
        intensity_bars=(v3, v1),
        intensity_target=v2,
    )

    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim = pad_dict_by_visual_size(stim, padding, ppd=ppd, pad_value=v2)
        params["padding"] = padding

    original_shape_np = SHAPES["white_yazdanbakhsh"]
    original_shape = np.array(original_shape_np) + 20
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["white_yazdanbakhsh"],
        original_range=(1, 9),
        intensity_range=(v3, v1),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def white_anderson(
    visual_size=VSIZES["white_anderson"], ppd=PPD, shape=SHAPES["white_anderson"], pad=PAD
):
    """Anderson variation of White stimulus, Domijan (2015) Fig 11B

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 10)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 100)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Anderson, B. L. (2001).
        Contrasting theories of White's illusion.
        Perception, 30, 1499-1501.
    Blakeslee, B., Pasieka, W., & McCourt, M. E. (2005).
        Oriented multiscale spatial ﬁltering and contrast normalization:
        a parsimonious model of brightness induction in a continuum
        of stimuli including White, Howe and simultaneous brightness contrast.
        Vision Research, 45, 607-615.
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(
        shape, visual_size, ppd, VSIZES["white_anderson"]
    )
    ppd = ppd[0]

    params = {
        "visual_size": visual_size,
        "ppd": ppd,
        "frequency": 5.0 / visual_size[1],
        "target_indices_top": (3,),
        "target_indices_bottom": (8,),
        "target_center_offset": visual_size[0] / 10.0,
        "target_height": visual_size[0] / 5.0,
        "stripe_center_offset": visual_size[0] / 5.0,
        "stripe_height": visual_size[0] / 5.0,
        "period": "even",
    }

    stim = stimupy.whites.anderson(
        **params,
        intensity_bars=(v1, v3),
        intensity_target=v2,
        intensity_stripes=(v1, v3),
    )

    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim = pad_dict_by_visual_size(stim, padding, ppd=ppd, pad_value=v2)
        params["padding"] = padding

    original_shape_np = SHAPES["white_anderson"]
    original_shape = np.array(original_shape_np) + 20
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["white_anderson"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


def white_howe(visual_size=VSIZES["white_howe"], ppd=PPD, shape=SHAPES["white_howe"], pad=PAD):
    """Howe variation of White stimulus, Domijan (2015) Fig 11C

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None
        visual size [height, width] in degrees, default: (10, 10)
    ppd : Sequence[Number, Number], Number, or None
        pixels per degree [vertical, horizontal], default: 10
    shape : Sequence[Number, Number], Number, or None
        shape [height, width] in pixels, default: (100, 100)
    pad : bool
        If True, include original padding (default: False)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., Pasieka, W., & McCourt, M. E. (2005).
        Oriented multiscale spatial ﬁltering and contrast normalization:
        a parsimonious model of brightness induction in a continuum
        of stimuli including White, Howe and simultaneous brightness contrast.
        Vision Research, 45, 607-615.
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    Howe, P. D. L. (2001).
        A comment on the Anderson (1997), the Todorovic (1997),
        and the Ross and Pessoa (2000) explanations of White's eﬀect.
        Perception, 30, 1023-1026
    """

    # Resolve resolution
    shape, visual_size, ppd, visual_resize = resolve(shape, visual_size, ppd, VSIZES["white_howe"])
    ppd = ppd[0]

    params = {
        "visual_size": visual_size,
        "ppd": ppd,
        "frequency": 5.0 / visual_size[1],
        "target_indices_top": (3,),
        "target_indices_bottom": (8,),
        "target_center_offset": visual_size[0] / 5.0,
        "target_height": visual_size[0] / 5.0,
        "period": "even",
    }

    stim = stimupy.whites.howe(
        **params,
        intensity_bars=(v1, v3),
        intensity_target=v2,
        intensity_stripes=(v1, v3),
    )

    if pad:
        padding = np.array((0.9, 1.1)) * visual_resize
        stim = pad_dict_by_visual_size(stim, padding, ppd=ppd, pad_value=v2)
        params["padding"] = padding

    original_shape_np = SHAPES["white_howe"]
    original_shape = np.array(original_shape_np) + 20
    original_visual_size = resolution.visual_size_from_shape_ppd(original_shape, PPD)
    params.update(
        original_shape=original_shape,
        original_ppd=PPD,
        original_visual_size=original_visual_size,
        original_shape_no_padding=original_shape_np,
        original_visual_size_no_padding=VSIZES["white_howe"],
        original_range=(1, 9),
        intensity_range=(v1, v3),
        visual_size=resolution.visual_size_from_shape_ppd(stim["img"].shape, ppd),
        shape=stim["img"].shape,
    )
    return {**stim, **params}


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = gen_all(skip=True)
    plot_stimuli(stims, mask=False)



import numpy as np

from stimupy.utils import pad_to_shape, resolution

__all__ = [
    "dungeon",
]


def dungeon(
    visual_size=None,
    ppd=None,
    shape=None,
    cell_size=None,
    n_cells=None,
    target_radius=1,
    intensity_background=0.0,
    intensity_grid=1.0,
    intensity_target=0.5,
):
    """Dungeon stimulus (Bressan, 2001) with diamond target.

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    cell_size : Sequence[Number, Number], Number, or None (default)
        size of individual cell (height, width)
    n_cells : Sequence[Number, Number], Number, or None (default)
        the number of square cells (not counting background) per dimension
    target_radius : int
        the "Manhattan radius" of the diamond target in # cells
    intensity_background : float
        intensity value for background
    intensity_grid : float
        intensity value for grid cells
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Bressan, P. (2001).
        Explaining lightness illusions.
        Perception, 30(9), 1031-1046.
        https://doi.org/10.1068/p3109
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    if isinstance(visual_size, (float, int)) or (visual_size is None):
        visual_size = (visual_size, visual_size)
    if isinstance(cell_size, (float, int)) or (cell_size is None):
        cell_size = (cell_size, cell_size)
    if isinstance(n_cells, (float, int)) or (n_cells is None):
        n_cells = (n_cells, n_cells)

    params = resolve_dungeon_params(None, visual_size, ppd, n_cells, cell_size)
    n_cells = (int(params["n_cells"][0]), int(params["n_cells"][1]))

    height, width = params["shape"].height, params["shape"].width
    cheight = height / (n_cells[0] * 2 - 1)
    cwidth = width / (n_cells[1] * 2 - 1)

    # create 2D array of grid
    arr = np.ones((n_cells[0] * 2 - 1, n_cells[1] * 2 - 1)) * intensity_grid
    mask_arr = np.zeros((n_cells[0] * 2 - 1, n_cells[1] * 2 - 1))

    # compute Manhattan distances from image center (=radius) of each pixel
    y = np.arange(-n_cells[0] + 1, n_cells[0])
    x = np.arange(-n_cells[1] + 1, n_cells[1])
    radii = np.abs(x[np.newaxis]) + np.abs(y[:, np.newaxis])

    # add targets
    idx = radii <= (target_radius * 2)
    arr[idx] = intensity_target
    mask_arr[idx] = 1

    # compute and apply grid mask
    grid_mask = [
        [(False if i % 2 == 0 and j % 2 == 0 else True) for i in range(n_cells[1] * 2 - 1)]
        for j in range(n_cells[0] * 2 - 1)
    ]
    grid_mask = np.array(grid_mask)
    arr[grid_mask] = intensity_background
    mask_arr[grid_mask] = 0

    img = arr.repeat(cheight, axis=0).repeat(cwidth, axis=1)
    mask = mask_arr.repeat(cheight, axis=0).repeat(cwidth, axis=1)

    # Make sure that stimulus size is as requested
    if (img.shape[0] != height) or (img.shape[1] != width):
        img = pad_to_shape(img, (height, width), intensity_background)
        mask = pad_to_shape(mask, (height, width), 0)

    stim = {
        "img": img,
        "target_mask": mask.astype(int),
        "target_radius": target_radius,
        "intensity_background": intensity_background,
        "intensity_grid": intensity_grid,
        "intensity_target": intensity_target,
        **params,
    }
    return stim


def resolve_dungeon_params(
    shape=None,
    visual_size=None,
    ppd=None,
    n_cells=None,
    cell_size=None,
):
    # Try to resolve resolution
    try:
        shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    except resolution.TooManyUnknownsError:
        ppd = resolution.validate_ppd(ppd)
        shape = resolution.validate_shape(shape)
        visual_size = resolution.validate_visual_size(visual_size)

    n_cells1, visual_angle1, cell_size1 = resolve_cells_1d(
        visual_size[0], n_cells[0] - 0.5, cell_size[0]
    )
    n_cells2, visual_angle2, cell_size2 = resolve_cells_1d(
        visual_size[1], n_cells[1] - 0.5, cell_size[1]
    )

    # Now resolve resolution
    shape, visual_size, ppd = resolution.resolve(
        shape=shape, visual_size=(visual_angle1 / 2, visual_angle2 / 2), ppd=ppd
    )

    return {
        "visual_size": visual_size,
        "ppd": ppd,
        "shape": shape,
        "n_cells": (n_cells1 + 0.5, n_cells2 + 0.5),
        "cell_size": (cell_size1 / 2, cell_size2 / 2),
    }


def resolve_cells_1d(
    visual_angle=None,
    n_cells=None,
    cell_size=None,
):
    # Try to resolve number and size of cells
    if cell_size is not None:
        cells_pd = 1 / cell_size / 2
    else:
        cells_pd = None

    try:
        n_cells, visual_angle, cells_pd = resolution.resolve_1D(
            length=n_cells, visual_angle=visual_angle, ppd=cells_pd
        )
        visual_angle = visual_angle * 2
        cell_size = 1 / cells_pd
    except Exception as e:
        raise Exception("Could not resolve n_cells, cell_size") from e

    return n_cells, visual_angle, cell_size


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 30,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "dungeon": dungeon(**default_params, n_cells=5)
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)
